/*
 * ESP32_BLE_KBD.ino
 *     Bluetooth Keyboard
 *
 *   ESP32   : MCU
 *   ILI9341 : LCD panel
 *   XPT2046 : Touch panel
 */ 
// ESP32_BLE_KBD.ino
//
 
// Satisfy the IDE, which needs to see the include statment in the ino too.
#ifdef dobogusinclude
#include <spi4teensy3.h>
#endif
#include "BleHidKeyboard.h"

#include <SPI.h>
#include <Adafruit_GFX.h>
#include <Adafruit_ILI9341.h>
#include <XPT2046_Touchscreen.h>

#define CS_PIN 17
#define TFT_DC  2
#define TFT_CS 15
#define TFT_RST 4

Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC, TFT_RST);

XPT2046_Touchscreen ts(CS_PIN);
#define TIRQ_PIN 16
#define TS_MINX 400
#define TS_MINY 300
#define TS_MAXX 3900
#define TS_MAXY 3800

#define XMAX 320
#define YMAX 240
#define SIZEX 29
#define SIZEY 32

#define LED   2
 
BleHidKeyboard* ble = 0;

boolean istouched, wastouched;

#define MAX_KEY 63

struct f_key_code {
  int code;
  int x;
  int y;
  int sx;
  int sy;
  char name[8];
  char shift[8];
} key_code[MAX_KEY] = {
  0x1E, 0, 0, 1, 1, "1", "!",
  0x1F, 1, 0, 1, 1, "2", "\"",
  0x20, 2, 0, 1, 1, "3", "#",
  0x21, 3, 0, 1, 1, "4", "$",
  0x22, 4, 0, 1, 1, "5", "%",
  0x23, 5, 0, 1, 1, "6", "&",
  0x24, 6, 0, 1, 1, "7", "'",
  0x25, 7, 0, 1, 1, "8", "(",
  0x26, 8, 0, 1, 1, "9", ")",
  0x27, 9, 0, 1, 1, "0", " ",
  0x2D,10, 0, 1, 1, "-", "=",
  
  0x14, 0, 1, 1, 1, "q", "Q",
  0x1A, 1, 1, 1, 1, "w", "W",
  0x08, 2, 1, 1, 1, "e", "E",
  0x15, 3, 1, 1, 1, "r", "R",
  0x17, 4, 1, 1, 1, "t", "T",
  0x1C, 5, 1, 1, 1, "y", "Y",
  0x18, 6, 1, 1, 1, "u", "U",
  0x0C, 7, 1, 1, 1, "i", "I",
  0x12, 8, 1, 1, 1, "o", "O",
  0x13, 9, 1, 1, 1, "p", "P",
  0x2F,10, 1, 1, 1, "@", "`",
  
  0x04, 0, 2, 1, 1, "a", "A",
  0x16, 1, 2, 1, 1, "s", "S",
  0x07, 2, 2, 1, 1, "d", "D",
  0x09, 3, 2, 1, 1, "f", "F",
  0x0A, 4, 2, 1, 1, "g", "G",
  0x0B, 5, 2, 1, 1, "h", "H",
  0x0D, 6, 2, 1, 1, "j", "J",
  0x0E, 7, 2, 1, 1, "k", "K",
  0x0F, 8, 2, 1, 1, "l", "L",
  
  0x1D, 0, 3, 1, 1, "z", "Z",
  0x1B, 1, 3, 1, 1, "x", "X",
  0x06, 2, 3, 1, 1, "c", "C",
  0x19, 3, 3, 1, 1, "v", "V",
  0x05, 4, 3, 1, 1, "b", "B",
  0x11, 5, 3, 1, 1, "n", "N",
  0x10, 6, 3, 1, 1, "m", "M",
  0x36, 7, 3, 1, 1, ",", "<",
  0x37, 8, 3, 1, 1, ".", ">",
  
  0x2E, 0, 4, 1, 1, "^", "~",
  0x89, 1, 4, 1, 1, "\\", "|",
  0x33, 2, 4, 1, 1, ";", "+",
  0x34, 3, 4, 1, 1, ":", "*",
  0x30, 4, 4, 1, 1, "[", "{",
  0x32, 5, 4, 1, 1, "]", "}",
  0x38, 6, 4, 1, 1, "/", "?",
  0x87, 7, 4, 1, 1, "\\", "_",

  0x49, 8, 4, 1, 1, "In", "In",
  0x4C, 8, 5, 1, 1, "De", "De",

  0x2A, 9, 2, 2, 1, "BS", "BS",
  0x28, 9, 3, 2, 2, "Ent", "Ent",
  0x29, 0, 5, 2, 1, "Esc", "Esc",
  0x35, 2, 5, 2, 1, "Chin", "Chin",
  0x2B, 4, 5, 2, 1, "Tab", "Tab",
//  0x39, 4, 5, 2, 1, "Caps", "Caps",
  0xE1, 6, 5, 2, 1, "Shft", "Shft",
  0xE0, 0, 6, 2, 1, "Ctrl", "Ctrl",
  0xE2, 2, 6, 2, 1, "Alt", "Alt",
  0x2C, 4, 6, 4, 1, "Space", "Space",
  0x52, 9, 5, 1, 1, "^", "^",
  0x51, 9, 6, 1, 1, "v", "v",
  0x50, 8, 6, 1, 1, "<", "<",
  0x4F,10, 6, 1, 1, ">", ">"
};

int key_alt;
int key_shft;
int key_ctrl;

void setup() {
  int i;
  
  pinMode(LED, OUTPUT);
  Serial.begin( 115200 );
  delay(100);
  Serial.println("ESP32_BLE_HID_KBD Start...");

  tft.begin();
  tft.setRotation(1);
  tft.fillScreen(ILI9341_BLACK);
  ts.begin();
  ts.setRotation(1);

  tft.setTextSize(2);
  tft.setTextColor(ILI9341_WHITE);
  tft.setCursor(0, YMAX/2-20);
  tft.print("Wait for");  
  tft.setCursor(20, YMAX/2);
  tft.print("Bluetooth connection !");  

  ble = BleHidKeyboard::getInstance();
  ble->set_led_pin(LED);
  delay( 200 );
  ble->init("ESP32_BLE_KBD");
  while(!ble->is_connected())
    delay(200);  

  key_alt=0;
  key_shft=0;
  key_ctrl=0;

  show_key(-1, key_alt, key_shft, key_ctrl);

  ble->report_release(0xE1, 0); // Release Shift
}
 
void loop() {
  struct TS_Point p;
  int mx, my;
  int i;
  char str[32];
  int  key;
  int  num;
  uint8_t modif;

  istouched=ts.touched();
  
  if (!istouched || wastouched){
    wastouched=istouched;
    delay(100);
    return;
  }

  wastouched=istouched;

  p=ts.getPoint();

  p.x = map(p.x, TS_MINX, TS_MAXX, 0, XMAX);
  p.y = map(p.y, TS_MINY, TS_MAXY, 0, YMAX);
  mx=XMAX-1-p.x;
  my=YMAX-1-p.y;

  key=-1;
  for(i=0;i<MAX_KEY;i++){
    if (key_code[i].x*SIZEX<=mx && key_code[i].y*SIZEY<=my &&
        mx<(key_code[i].x+key_code[i].sx)*SIZEX && my<(key_code[i].y+key_code[i].sy)*SIZEY){
          num=i;
          key=key_code[num].code;
          break;
    }
  }

  if (key>=0){
    show_key(num, key_alt, key_shft, key_ctrl);
    // Shift
    if (key==0xE1 || key==0xE5){
      if (key_shft==0){
        key_shft=1;
        sprintf(str, "DOWN : %02X", key);
        Serial.println(str);
        modif=key_ctrl+key_shft*2;
        ble->report_press(key, modif);
      }else{
        key_shft=0;
        sprintf(str, "UP : %02X", key);
        Serial.println(str);
        modif=key_ctrl+key_shft*2;
        ble->report_release(key, modif);
      }
    
    // Ctrl
    }else if (key==0xE0 || key==0xE4){
      modif=key_ctrl+key_shft*2;
      if (key_ctrl==0){
        key_ctrl=1;
        sprintf(str, "DOWN : %02X", key);
        Serial.println(str);
        modif=key_ctrl+key_shft*2;
        ble->report_press(key, modif);
      }else{
        key_ctrl=0;
        sprintf(str, "UP : %02X", key);
        Serial.println(str);
        modif=key_ctrl+key_shft*2;
        ble->report_release(key, modif);
      }

    // Alt
    }else if (key==0xE2 || key==0xE6){
      modif=key_ctrl+key_shft*2+key_alt*4;
      if (key_alt==0){
        key_alt=1;
        sprintf(str, "DOWN : %02X", key);
        Serial.println(str);
        modif=key_ctrl+key_shft*2+key_alt*4;
        ble->report_press(key, modif);
      }else{
        key_alt=0;
        sprintf(str, "UP : %02X", key);
        Serial.println(str);
        modif=key_ctrl+key_shft*2+key_alt*4;
        ble->report_release(key, modif);
      }

    }else{
      modif=key_ctrl+key_shft*2+key_alt*4;
      sprintf(str, "DOWN : %02X", key);
      Serial.println(str);
      ble->report_press(key, modif);
      delay(10);
      sprintf(str, "UP : %02X", key);
      Serial.println(str);
      ble->report_release(key, modif);
    }
    show_key(-1, key_alt, key_shft, key_ctrl);
  }

  delay(10);
}

void show_key(int key_num, int alt, int shft, int ctrl)
{
  int i;
  
  tft.fillScreen(ILI9341_BLACK);
  tft.setTextSize(2);
  tft.setTextColor(ILI9341_WHITE);

  for(i=0;i<MAX_KEY;i++){
    if (shft==1 && key_code[i].code==0xE1){
      tft.fillRect(key_code[i].x*SIZEX, key_code[i].y*SIZEY, key_code[i].sx*SIZEX, key_code[i].sy*SIZEY, ILI9341_RED);
    }
    if (ctrl==1 && key_code[i].code==0xE0){
      tft.fillRect(key_code[i].x*SIZEX, key_code[i].y*SIZEY, key_code[i].sx*SIZEX, key_code[i].sy*SIZEY, ILI9341_RED);
    }
    if (alt==1 && key_code[i].code==0xE2){
      tft.fillRect(key_code[i].x*SIZEX, key_code[i].y*SIZEY, key_code[i].sx*SIZEX, key_code[i].sy*SIZEY, ILI9341_RED);
    }
    if (i==key_num)
      tft.fillRect(key_code[i].x*SIZEX, key_code[i].y*SIZEY, key_code[i].sx*SIZEX, key_code[i].sy*SIZEY, ILI9341_YELLOW);
    tft.drawRect(key_code[i].x*SIZEX, key_code[i].y*SIZEY, key_code[i].sx*SIZEX, key_code[i].sy*SIZEY, ILI9341_WHITE);
    tft.setCursor(key_code[i].x*SIZEX+6, key_code[i].y*SIZEY+4);

    if (shft==1)
      tft.print(key_code[i].shift);
    else
      tft.print(key_code[i].name);
  }
}
